from django.urls import include, path

from .views import EditorAllSectionsView, \
    GenericEntitiesCreateView, GenericEntitiesUpdateView, GenericItemCreateView, get_custom_ajax_form, \
    save_custom_ajax_form, CustomTableActionButtonView, ds_select_post, set_start_page, update_browser, \
    editor_session_update, \
    UpdatePageTokenView, editor_update_section, EditorPageView, editor_autocomplete, ProjectPageView, \
    update_datasource, upload_pdf, create_import_project, edit_import_project, update_datasource_select_ajax

from django.conf.urls import url

app_name = 'apis_import_project'

editor_patterns = [
    url(r'^(?P<project_pk>[0-9]+)/$', EditorPageView.as_view(), name='editor_main'),
    url(r'^session/update$', editor_session_update, name='editor_session_update'),
    url(r'^section/all/$',
        EditorAllSectionsView.as_view(),
        name='get_sections', ),
    url(r'^section/update/(?P<section>[A-Za-z]+)/$', editor_update_section, name='editor_section_update'),
    url(r'^autocomplete/(?P<entity>[a-zA-Z]+)/$', editor_autocomplete, name='editor_autocomplete'),

    url(r'^^entity/create/(?P<entity>[a-zA-Z]+)/$',
        GenericEntitiesCreateView.as_view(),
        name='entity_create', ),
    url(r'^^entity/update/(?P<entity>[a-zA-Z]+)/(?P<pk>[0-9]+)/$',
        GenericEntitiesUpdateView.as_view(),
        name='entity_update', ),
    url(r'^item/create/(?P<model_name>[A-Za-z]+)$',
        GenericItemCreateView.as_view(),
        name='item_create', ),

    url(r'^relation/getform/$', get_custom_ajax_form, name='relation_get_form'),
    url(r'^relation/autodate/(?P<rel_pk>[0-9]+)/(?P<tab>[a-zA-Z_]+)/$', CustomTableActionButtonView.as_view(),
        name='relation_autodate'),
    url(
            r'^relation/update/(?P<col_pk>[0-9]+)/(?P<pagedata_pk>[0-9]+)/(?P<entity_type>[a-zA-z]+)/(?P<kind_form>['
            r'a-zA-z]+)/(?P<SiteID>[0-9]+)(?:/(?P<ObjectID>[0-9]*))?/$',
            save_custom_ajax_form, name='relation_update_or_create'
    ),
    url(
            r'^relation/create/(?P<entity_type>[a-zA-z]+)/(?P<kind_form>[a-zA-z]+)/(?P<SiteID>[0-9]+)(?:/(?P<ObjectID>['
            r'0-9]*))?/$',
            save_custom_ajax_form, name='relation_update_or_create'
    ),

]

datasource_patterns = [
    url(r'^select/$', ds_select_post, name='ds_select'),
    url(r'^upload/$', upload_pdf, name='ds_upload'),
    url(r'^page/update/$', UpdatePageTokenView.as_view(), name='page_set_token'),
    url(r'update/$', update_datasource, name='ds_update'),
    url(r'^update_select/$', update_datasource_select_ajax, name='ds_list_refresh'),
    url(r'^page/update/start$', set_start_page, name='page_set_start'),

]

browser_patterns = [
    url(r'^refresh/$', update_browser,
        name='browser_refresh'),
    url(r'^refresh/(?P<project_pk>[0-9]+)/(?P<data_source_pk>[0-9]+)/(?P<page_num>[0-9]+)/(?P<direction>[a-z]+)/$',
        update_browser,
        name='browser_refresh'),
]

project_patterns = [
    path('', ProjectPageView.as_view(), name='project_main'),
    url(r'^create/$', create_import_project, name='project_create'),
    url(r'^update/(?P<project_pk>[0-9]+)/$', edit_import_project, name='project_update'),
]

urlpatterns = [
    path('editor/', include(editor_patterns)),
    path('datasource/', include(datasource_patterns)),
    path('project/', include(project_patterns)),
    path('browser/', include(browser_patterns)),
]



import re

from django.conf import settings
from django.db.models.query import QuerySet
from django.urls import reverse
from rest_framework import serializers
from apis_core.apis_labels.serializers import LabelSerializerLegacy as LabelSerializer
from apis_core.apis_entities.serializers_generic import VocabsSerializer

base_uri = getattr(settings, "APIS_BASE_URI", "http://apis.info")
if base_uri.endswith("/"):
    base_uri = base_uri[:-1]

class EntitySerializer(serializers.Serializer):
    id = serializers.IntegerField()
    name = serializers.CharField()
    start_date = serializers.DateField()
    end_date = serializers.DateField()
    labels = LabelSerializer(source="label_set", many=True)

    def add_entity_type(self, obj):
        return str(obj.__class__.__name__)


    def __init__(
        self, *args, **kwargs
    ):
        super(EntitySerializer, self).__init__(*args, **kwargs)
        if type(self.instance) == QuerySet:
            inst = self.instance[0]
        else:
            inst = self.instance
        if inst is None:
            return
        for f in inst._meta.fields:
            field_name = re.search(r"([A-Za-z]+)\'>", str(f.__class__)).group(1)
            if field_name in [
                "CharField",
                "DateField",
                "DateTimeField",
                "IntegerField",
                "FloatField",
            ]:
                self.fields[f.name] = getattr(serializers, field_name)()
            elif field_name in ["ForeignKey", "ManyToMany"]:
                if str(f.related_model.__module__).endswith("apis_vocabularies.models"):
                    many = False
                    if f.many_to_many or f.one_to_many:
                        many = True
                    self.fields[f.name] = VocabsSerializer(many=many)
        for f in inst._meta.many_to_many:
            if f.name.endswith("relationtype_set"):
                continue
            elif f.name == "collection":
                pass
            elif str(f.related_model.__module__).endswith("apis_vocabularies.models"):
                self.fields[f.name] = VocabsSerializer(many=True)
        self.fields["entity_type"] = serializers.SerializerMethodField(
            method_name="add_entity_type"
        )



class RelationEntitySerializer(serializers.Serializer):
    id = serializers.IntegerField()
    start_date = serializers.DateField()
    end_date = serializers.DateField()
    start_date_written = serializers.DateField()
    end_date_written = serializers.DateField()
    relation_type = serializers.SerializerMethodField(method_name="add_relation_label")



    def add_entity(self, obj):
        return EntitySerializer(
            getattr(obj, "related_{}".format(self.entity_type)), depth_ent=0
        ).data

    def add_relation_label(self, obj):
        cm = obj.__class__.__name__
        res_1 = dict()
        res_1["id"] = obj.relation_type.pk
        res_1[
            "url"
        ] = f"{base_uri}{reverse('apis_core:apis_api:{}relation-detail'.format(cm).lower(), kwargs={'pk': obj.relation_type.pk},)}"
        if self.reverse and len(obj.relation_type.label_reverse) > 0:
            res_1["label"] = obj.relation_type.label_reverse
        elif self.reverse:
            res_1["label"] = "({})".format(obj.relation_type.label)
        else:
            res_1["label"] = obj.relation_type.label
        return res_1

    def __init__(self, *args, own_class=None, reverse=False, **kwargs):
        super(RelationEntitySerializer, self).__init__(*args, **kwargs)
        self.own_class = own_class
        self.reverse = reverse
        if self.instance is not None:
            for f in self.instance._meta.fields:
                if f.name.startswith("related_"):
                    mk2 = f.name.replace("related_", "")

                    if mk2.lower() != own_class.lower():
                        self.entity_type = mk2
                        self.fields["target"] = serializers.SerializerMethodField(
                            method_name="add_entity"
                        )



# -*- coding: utf-8 -*-
import copy
import logging
from crispy_forms.layout import Layout, Fieldset, Submit, Button, Field
from crispy_forms.bootstrap import Accordion, AccordionGroup
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Fieldset, Button, Submit
from dal import autocomplete
import re
import sys
import inspect
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
from django.db.models import Q
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.validators import URLValidator
from django.forms import ModelMultipleChoiceField, ModelChoiceField, HiddenInput
from django.urls import reverse
from django import forms

from apis_core.apis_metainfo.models import TempEntityClass, Text, Uri, Collection
from apis_core.apis_relations.forms2 import validate_target_autocomplete
from apis_core.apis_relations.models import AbstractRelation
from apis_core.apis_vocabularies.models import VocabsBaseClass, AbstractRelationType
from apis_core.apis_labels.forms import LabelForm
from apis_import_project.tables import get_generic_relations_table

if 'apis_highlighter' in settings.INSTALLED_APPS:
    from apis_highlighter.models import Annotation, AnnotationProject

from apis_core.helper_functions import DateParser
from apis_core.helper_functions.RDFParser import RDFParser
from apis_core.apis_entities.fields import ListSelect2, Select2Multiple
from apis_core.apis_entities.models import AbstractEntity
from .models import DataSource, ImportProject

log = logging.getLogger("mylogger")
log.setLevel(logging.WARNING)

class EditDataSourceForm(forms.ModelForm):
    class Meta:
        model = DataSource

        fields = ["name", "citation", "year", "server_directory", "page_count"]

    def __init__(self, *args, **kwargs):
        super(EditDataSourceForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.from_tag = False
        self.helper.form_class = "update_datasource_form"
        self.helper.layout = Layout("name", "citation", "year", "server_directory", "page_count")
        self.helper.form_method = "POST"
        self.helper.add_input(Submit("submit", "save", css_id="post_update_datasource"))
        self.helper.add_input(Button('cancel', 'Cancel', css_class='btn-danger',
                                     onclick="updateBrowserPage()"))


class CreateImportProjectForm(forms.ModelForm):
    class Meta:
        model = ImportProject
        fields = ["name", "description", "Editors", "DataSources", "owner"]

    def __init__(self, user=None, *args, **kwargs):
        super(CreateImportProjectForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()

        ds = DataSource.objects.filter(owner=user)
        ed = User.objects.exclude(pk=user.pk)
        self.fields["owner"].widget = HiddenInput()
        self.fields["owner"].initial = user.pk

        if ds:
            self.fields["DataSources"].queryset = ds
            self.helper.layout = Layout("name", "description", "Editors", "DataSources", "owner")
        else:
            del self.fields["DataSources"]
            self.helper.layout = Layout("name", "description", "Editors", "owner")

        if ed:
            self.fields["Editors"].queryset = ed
            self.helper.layout = Layout("name", "description", "Editors", "DataSources", "owner")
        else:
            del self.fields["Editors"]
            self.helper.layout = Layout("name", "description", "DataSources", "owner")

        # self.helper.form_tag = False
        self.helper.form_method = "POST"
        self.helper.add_input(Submit("submit", "Submit"))
        self.helper.add_input(Button('cancel', 'Cancel', css_class='btn-danger',
                                     onclick="window.location.href = '{}';".format(
                                             reverse('apis_import_project:project_main'))))


class UploadFileForm(forms.Form):
    """
    Combines two models:
    names, citation and year are fields in DataSource
    file is processed and then split into DataSourcePageObjects
    """
    name = forms.CharField(max_length=50)
    citation = forms.CharField(max_length=500)
    year = forms.CharField(max_length=10)
    server_directory = forms.CharField(max_length=255)
    page_count = forms.IntegerField(max_value=10000)

    def __init__(self, *args, project_pk, **kwargs):
        super(UploadFileForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.from_tag = False
        self.helper.layout = Layout("name", "citation", "year", "server_directory", "page_count")
        self.helper.add_input(Submit("submit", "Submit"))
        self.helper.add_input(Button('cancel', 'Cancel', css_class='btn-danger',
                                     onclick="updateBrowserPage(null, null)"))


def get_custom_generic_entities_form(entity):
    # fixme: __maintainence__ __gpirgie__ Hook up with existing apis core function/class if possible

    class CustomGenericEntitiesForm(forms.ModelForm):
        class Meta:
            model = AbstractEntity.get_entity_class_of_name(entity)

            exclude = [
                "start_date",
                "start_start_date",
                "start_end_date",
                "start_date_is_exact",
                "end_date",
                "end_start_date",
                "end_end_date",
                "end_date_is_exact",
                "text",
                "source",
                "published",
            ]
            exclude.extend(model.get_related_entity_field_names())
            exclude.extend(model.get_related_relationtype_field_names())
            if entity == "Person":
                exclude += ["profession", "review", "status"]
            else:
                exclude += ["review", "status"]

        def __init__(self, *args, **kwargs):
            super(CustomGenericEntitiesForm, self).__init__(*args, **kwargs)
            self.helper = FormHelper()
            self.helper.form_class = entity.title() + "Form"
            self.helper.form_tag = False
            self.helper.help_text_inline = True
            acc_grp1 = Fieldset("Metadata {}".format(entity.title()))
            acc_grp2 = []  # AccordionGroup("MetaInfo", "references", "notes")#, "review")
            attrs = {
                "data-placeholder": "Type to get suggestions",
                "data-minimum-input-length": getattr(settings, "APIS_MIN_CHAR", 3),
                "data-html": True,
            }

            # list to catch all fields that will not be inserted into accordion group acc_grp2
            fields_list_unsorted = []

            for f in self.fields.keys():
                if isinstance(
                        self.fields[f], (ModelMultipleChoiceField, ModelChoiceField)
                ):
                    v_name_p = str(self.fields[f].queryset.model.__name__)
                    if isinstance(self.fields[f], ModelMultipleChoiceField):
                        widget1 = Select2Multiple
                    else:
                        widget1 = ListSelect2
                    if (
                            ContentType.objects.get(
                                    app_label__in=[
                                        "apis_entities",
                                        "apis_metainfo",
                                        "apis_relations",
                                        "apis_vocabularies",
                                        "apis_labels",
                                    ],
                                    model=v_name_p.lower(),
                            ).app_label.lower()
                            == "apis_vocabularies"
                    ):
                        self.fields[f].widget = widget1(
                                url=reverse(
                                        "apis:apis_vocabularies:generic_vocabularies_autocomplete",
                                        kwargs={"vocab": v_name_p.lower(), "direct": "normal"},
                                ),
                                attrs=attrs,
                        )
                        if self.instance:
                            res = []
                            if isinstance(self.fields[f], ModelMultipleChoiceField):
                                try:
                                    for x in getattr(self.instance, f).all():
                                        res.append((x.pk, x.label))
                                except ValueError:
                                    pass
                                self.fields[f].initial = res
                                self.fields[f].choices = res
                            else:
                                try:
                                    res = getattr(self.instance, f)
                                    if res is not None:
                                        self.fields[f].initial = (res.pk, res.label)
                                        self.fields[f].choices = [
                                            (res.pk, res.label),
                                        ]
                                except ValueError:
                                    res = ""
                if f not in acc_grp2:
                    # append to unsorted list, so that it can be sorted and afterwards attached to accordion group
                    # acc_grp1
                    fields_list_unsorted.append(f)

            def sort_fields_list(list_unsorted, entity_label):
                """
                Sorts a list of model fields according to a defined order.


                :param list_unsorted: list
                    The unsorted list of fields.

                :param entity_label: str
                    The string representation of entity type, necessary to find the entity-specific ordering (if it
                    is defined)


                :return: list
                    The sorted list if entity-specific ordering was defined, the same unordered list if not.
                """

                sort_dic = {
                    "Person": ["first_name", "name", "start_date_written", "end_date_written", "gender",
                               "title", "collection", "references", "notes"],
                    "Place": ["name", "kind", "lat", "lng", "start_date_written", "end_date_written", "references",
                              "notes"],
                    "Institution": ["name", "kind", "start_date_written", "end_date_written", "references", "notes"],
                    "Work": ["name", "kind", "start_date_written", "end_date_written", "references", "notes"],
                    "Event": ["name", "kind", "start_date_written", "end_date_written", "references", "notes"],
                }

                sort_preferences = sort_dic.get(entity_label)

                sort_preferences_used = []

                if sort_preferences is None:
                    return list_unsorted
                else:
                    # list of tuples to be sorted later
                    field_rank_pair_list = []
                    for field in list_unsorted:
                        if field in sort_preferences:
                            # if this succeeds, then the field has been given a priorites ordering above
                            ranking_by_index = sort_preferences.index(field)
                            sort_preferences_used.append(field)
                            field_rank_pair = (field, ranking_by_index)
                        else:
                            # if no ordering for the field was found, then give it 'Inf'
                            # so that it will be attached at the end.
                            field_rank_pair = (field, float('Inf'))
                        field_rank_pair_list.append(field_rank_pair)
                    # Make a check if all items of sort_preferences were used. If not, this indicates an out of sync
                    # setting
                    # if len(sort_preferences) > 0:
                    if len(sort_preferences_used) != len(sort_preferences):

                        differences = []
                        for p in sort_preferences_used:
                            if p not in sort_preferences:
                                differences.append(p)
                        for p in sort_preferences:
                            if p not in sort_preferences_used:
                                differences.append(p)

                        raise Exception(
                                "An item of the entity setting 'form_order' list was not used. \n"
                                "This propably indicates that the 'form_order' settings is out of sync with the "
                                "effective django models.\n"
                                f"The relevant entity is: {entity_label}\n"
                                f"And the differences between used list and settings list are: {differences}"
                        )
                    # sort the list according to the second element in each tuple
                    # and then take the first elements from it and return as list
                    return [t[0] for t in sorted(field_rank_pair_list, key=lambda x: x[1])]

            # sort field list, iterate over it and append each element to the accordion group

            for f in sort_fields_list(fields_list_unsorted, entity):
                acc_grp1.append(f)

            self.helper.layout = Layout(Accordion(acc_grp1))  # , acc_grp2))
            if self.fields.get("status"):
                self.fields["status"].required = False
            self.fields["collection"].required = False
            self.fields["start_date_written"].required = False
            self.fields["end_date_written"].required = False
            self.fields["name"].required = True

            instance = getattr(self, "instance", None)
            if instance != None:

                if instance.start_date_written:
                    self.fields[
                        "start_date_written"
                    ].help_text = DateParser.get_date_help_text_from_dates(
                            single_date=instance.start_date,
                            single_start_date=instance.start_start_date,
                            single_end_date=instance.start_end_date,
                            single_date_written=instance.start_date_written,
                    )
                else:
                    self.fields[
                        "start_date_written"
                    ].help_text = DateParser.get_date_help_text_default()

                if instance.end_date_written:
                    self.fields[
                        "end_date_written"
                    ].help_text = DateParser.get_date_help_text_from_dates(
                            single_date=instance.end_date,
                            single_start_date=instance.end_start_date,
                            single_end_date=instance.end_end_date,
                            single_date_written=instance.end_date_written,
                    )
                else:
                    self.fields[
                        "end_date_written"
                    ].help_text = DateParser.get_date_help_text_default()

        def save(self, *args, **kwargs):
            obj = super(CustomGenericEntitiesForm, self).save(*args, **kwargs)
            if obj.collection.all().count() == 0:
                col_name = getattr(
                        settings, "APIS_DEFAULT_COLLECTION", "manually created entity"
                )
                col, created = Collection.objects.get_or_create(name=col_name)
                obj.collection.add(col)
            return obj

    return CustomGenericEntitiesForm



class GenericEntitiesStanbolForm(forms.Form):
    # fixme: __maintainence__ __gpirgie__ Hook up with existing apis core function/class if possible

    def save(self, *args, **kwargs):
        cd = self.cleaned_data
        entity = RDFParser(cd["entity"], self.entity.title()).get_or_create()
        return entity

    def __init__(self, entity, *args, **kwargs):

        attrs = {
            "data-placeholder": "Type to get suggestions",
            "data-minimum-input-length": getattr(settings, "APIS_MIN_CHAR", 3),
            "data-html": True,
            "style": "width: auto",
        }
        ent_merge_pk = kwargs.pop("ent_merge_pk", False)
        super(GenericEntitiesStanbolForm, self).__init__(*args, **kwargs)
        self.entity = entity
        self.helper = FormHelper()
        form_kwargs = {"entity": entity}
        url = reverse(
            "apis:apis_entities:generic_entities_autocomplete",
            args=[entity.title(), "remove"],
        )
        label = "Create {} from reference resources".format(entity.title())
        button_label = "Create"
        if ent_merge_pk:
            form_kwargs["ent_merge_pk"] = ent_merge_pk
            url = reverse(
                "apis:apis_entities:generic_entities_autocomplete",
                args=[entity.title(), ent_merge_pk],
            )
            label = "Search for {0} in reference resources or db".format(entity.title())
            button_label = "Merge"
        self.helper.form_action = reverse(
            "apis:apis_entities:generic_entities_stanbol_create", kwargs=form_kwargs
        )
        self.helper.add_input(Submit("submit", button_label))
        self.fields["entity"] = autocomplete.Select2ListCreateChoiceField(
            label=label,
            widget=ListSelect2(url=url, attrs=attrs),
            validators=[URLValidator],
        )


############### Create Forms Generic Section ####################
def get_subclass_dic(superclass, module):
    classes_in_module = [(name, object) for name, object in inspect.getmembers(module, inspect.isclass)]
    subclass_dic = {key: val for key, val in classes_in_module if superclass.__subclasscheck__(val)}

    return subclass_dic


def get_model_class_of_name(superclass, module, model_name):
    module = sys.modules[module]
    subclass_dic = get_subclass_dic(superclass, module)
    res = subclass_dic[model_name]

    return res


class GenericLabelForm(LabelForm):
    pass


#
#     def __init__(self, *args, **kwargs):
#         super(GenericLabelForm, self).__init__(*args, **kwargs)
#         self.helper.form_tag = True
#         self.helper.form_method = "post"
#         self.helper.form_action = f"/apis_import_project/create_item/Label"

def get_generic_vocabs_form(model_name):
    module = "apis_core.apis_vocabularies.models"
    "Returns a form for all models ending with -Type (i.e. 'TextType', 'CollectionType', 'InstitutionType'), " \
    "plus also the 'Title' model."

    class GenericVocabsForm(forms.ModelForm):
        class Meta:
            model = get_model_class_of_name(VocabsBaseClass, module, model_name)
            fields = ["name"]  # todo: __gpirgie__ check if more fields should be added

        def __init__(self, *args, **kwargs):
            super(GenericVocabsForm, self).__init__(*args, **kwargs)
            self.helper = FormHelper()
            self.helper.add_input(Submit("submit", "save", css_class=f"{model_name}Form"))
            self.helper.form_method = "post"
            self.helper.form_class = "form pdf_tool_itemcreate"
            self.helper.form_action = reverse("apis_import_project:item_create", kwargs={"model_name": model_name})  
            self.helper.layout = Layout(Field("name",
                                              id=f"{model_name}_name_field"))  # NOTE: used to clean the fields in
            # Ajax success function.

    return GenericVocabsForm


def get_generic_relationtype_form(model_name):
    "Returns a form for all models ending with -Relation (i.e. 'PersonInstitutionRelation', etc.)"
    module = "apis_core.apis_vocabularies.models"

    class GenericRelationTypeForm(forms.ModelForm):
        class Meta:
            model = get_model_class_of_name(AbstractRelationType, module, model_name)
            fields = ["name", "name_reverse"]  # todo: check if more fields should be added

        def __init__(self, *args, **kwargs):
            super(GenericRelationTypeForm, self).__init__(*args, **kwargs)
            self.helper = FormHelper()
            self.helper.add_input(Submit("submit", "save", css_class=f"{model_name}Form"))
            self.helper.form_method = "post"
            self.helper.form_class = "form pdf_tool_itemcreate"
            self.helper.form_action = reverse("apis_import_project:item_create", kwargs={"model_name": model_name})
            self.helper.layout = Layout(Field("name", id=f"{model_name}Form_name_field"),
                                        Field("name_reverse", id=f"{model_name}Form_name_reverse_field"))
            self.fields["name_reverse"].required = False


    return GenericRelationTypeForm


def get_generic_entity_type_form(model_name):
    module = "apis_core.apis_vocabularies.models"

    class GenericEntityTypeForm(forms.ModelForm):
        class Meta:
            model = get_model_class_of_name(VocabsBaseClass, module, model_name)
            fields = ["name", "description"]  # todo: check if more fields should be added

        def __init__(self, *args, **kwargs):
            super(GenericEntityTypeForm, self).__init__(*args, **kwargs)
            self.helper = FormHelper()
            self.helper.add_input(Submit("submit", "save", css_class=f"{model_name}Form"))
            self.helper.form_method = "post"
            self.helper.form_class = "form pdf_tool_itemcreate"
            self.helper.form_action = reverse("apis_import_project:item_create", kwargs={"model_name": model_name})
            self.helper.layout = Layout(Field("name", id=f"{model_name}Form_name_field"),
                                        Field("description", id=f"{model_name}Form_description_field"))
            self.fields["description"].required = False

    return GenericEntityTypeForm


class CustomGenericRelationForm(forms.ModelForm):
    # fixme: __maintainence__ __gpirgie__ Hook up with existing apis core function/class if possible

    class Meta:
        model = TempEntityClass
        fields = ['start_date_written', 'end_date_written', 'references', 'notes']
        labels = {
            'start_date_written': _('Start'),
            'end_date_written': _('End'),
        }

    def save(self, site_instance, instance=None, commit=True):
        """
        Save function of the GenericRelationForm.
        :param site_instance: Instance where the form is used on
        :param instance: PK of the relation that is saved
        :param commit: Whether to already commit the save.
        :type site_instance: object
        :type instance: int
        :type commit: bool
        :rtype: object
        :return: instance of relation
        """
        cd = self.cleaned_data
        if instance:
            x = self.relation_form.objects.get(pk=instance)
        else:
            x = self.relation_form()
        x.relation_type_id = cd['relation_type']
        x.start_date_written = cd['start_date_written']
        x.end_date_written = cd['end_date_written']
        x.notes = cd['notes']
        x.references = cd['references']
        setattr(x, self.rel_accessor[3], site_instance)
        target = AbstractEntity.get_entity_class_of_name(self.rel_accessor[0])
        t1 = target.get_or_create_uri(cd['target'])
        if not t1:
            t1 = RDFParser(cd['target'], self.rel_accessor[0]).get_or_create()
        setattr(x, self.rel_accessor[2], t1)
        if self.highlighter:
            an_proj = AnnotationProject.objects.get(pk=int(self.request.session.get('annotation_project', 1)))
            x.published = an_proj.published
        if commit:
            x.save()
        if self.highlighter:
            if not commit:
                x.save()
            txt = Text.objects.get(pk=cd['HL_text_id'][5:])
            a = Annotation(
                    start=cd['HL_start'],
                    end=cd['HL_end'],
                    text=txt,
                    user_added=self.request.user,
                    annotation_project_id=int(self.request.session.get('annotation_project', 1)))
            a.entity_link = x
            a.save()
        print('saved: {}'.format(x))
        return x

    def get_text_id(self):
        """
        Function to retrieve the highlighted text.
        :return: ID of text that was highlighted
        """
        return self.cleaned_data['HL_text_id'][5:]

    def get_html_table(self, entity_type, request, site_instance, form_match):
        table = get_generic_relations_table(relation_class=self.relation_form, entity_instance=site_instance,
                                            detail=False)
        prefix = re.match(r'([A-Z][a-z])[^A-Z]*([A-Z][a-z])', self.relation_form.__name__)
        prefix = prefix.group(1) + prefix.group(2) + '-'
        if form_match.group(1) == form_match.group(2):
            dic_a = {'related_' + entity_type.lower() + 'A': site_instance}
            dic_b = {'related_' + entity_type.lower() + 'B': site_instance}
            if 'apis_highlighter' in settings.INSTALLED_APPS:
                objects = self.relation_form.objects.filter_ann_proj(request=request).filter(
                        Q(**dic_a) | Q(**dic_b)
                )
            else:
                objects = self.relation_form.objects.filter(
                        Q(**dic_a) | Q(**dic_b)
                )

            table_html = table(data=objects, prefix=prefix)
        else:
            tab_query = {'related_' + entity_type.lower(): site_instance}
            if 'apis_highlighter' in settings.INSTALLED_APPS:
                ttab = self.relation_form.objects.filter_ann_proj(
                        request=request).filter(**tab_query)
            else:
                ttab = self.relation_form.objects.filter(**tab_query)
            table_html = table(data=ttab, prefix=prefix)
        return table_html

    def __init__(self, siteID=None, highlighter=False, prefill_inst=None, prefill_func=None, prefill_notes=None, *args,
                 **kwargs):
        """
        Generic Form for relations.
        :param siteID: ID of the entity the form is used on
        :param entity_type: Entity type of the entity the form is used on
        :param relation_form: Type of relation form.
        :param instance: instance of relation.
        :param highlighter: whether the form is used in the highlighter
        :type siteID: int
        :type entity_type: object or int
        :type relation_form: object or int
        :type instance: object
        :type highlighter: bool
        """

        attrs = {
            'data-placeholder': 'Type to get suggestions',
            'data-minimum-input-length': getattr(settings, "APIS_MIN_CHAR", 3),
            'data-html': True,
            'style': 'width: 100%'
        }
        help_text_target = "Search and select or use an URL from a reference resource"
        attrs_target = copy.deepcopy(attrs)
        attrs_target['data-tags'] = '1'
        css_notes = 'LS'
        self.highlighter = highlighter
        entity_type = kwargs.pop('entity_type')
        if type(entity_type) != str:
            entity_type = entity_type.__name__
        self.relation_form = kwargs.pop('relation_form')
        if type(self.relation_form) == str:
            self.relation_form = AbstractRelation.get_relation_class_of_name(self.relation_form)
        self.request = kwargs.pop('request', False)
        super(CustomGenericRelationForm, self).__init__(*args, **kwargs)
        instance = getattr(self, 'instance', None)
        self.fields['relation_type'] = forms.CharField(label='Relation type', required=True)
        self.helper = FormHelper()
        self.helper.form_class = '{}Form'.format(str(self.relation_form))
        self.helper.form_tag = False
        lst_src_target = re.findall('[A-Z][^A-Z]*', self.relation_form.__name__)
        if lst_src_target[0] == lst_src_target[1]:
            if instance and instance.id:
                if getattr(instance, 'related_{}A_id'.format(lst_src_target[0].lower())) == int(siteID):
                    self.rel_accessor = (lst_src_target[1], True,
                                         'related_{}B'.format(lst_src_target[1].lower()),
                                         'related_{}A'.format(lst_src_target[0].lower()))
                else:
                    self.rel_accessor = (lst_src_target[1], False,
                                         'related_{}A'.format(lst_src_target[1].lower()),
                                         'related_{}B'.format(lst_src_target[0].lower()))
            else:
                self.rel_accessor = (lst_src_target[1], True,
                                     'related_{}B'.format(lst_src_target[1].lower()),
                                     'related_{}A'.format(lst_src_target[0].lower()))
            self.fields['relation_type'] = autocomplete.Select2ListCreateChoiceField(
                    label='Relation type',
                    widget=ListSelect2(
                            # url='/vocabularies/autocomplete/{}{}relation/normal'.format(lst_src_target[0].lower(),
                            # lst_src_target[1].lower()),
                            url=reverse('apis:apis_vocabularies:generic_vocabularies_autocomplete', args=[
                                ''.join([lst_src_target[0].lower(), lst_src_target[1].lower(), 'relation']), 'normal']),
                            attrs=attrs))
            self.fields['target'] = autocomplete.Select2ListCreateChoiceField(
                    label=lst_src_target[1],
                    widget=ListSelect2(
                            # url='/entities/autocomplete/{}'.format(lst_src_target[1].lower()),
                            url=reverse('apis:apis_entities:generic_entities_autocomplete',
                                        args=[lst_src_target[1].lower()]),
                            attrs=attrs_target),
                    validators=[validate_target_autocomplete],
                    help_text=help_text_target)
        elif entity_type.lower() == lst_src_target[0].lower():
            self.rel_accessor = (lst_src_target[1], True,
                                 'related_{}'.format(lst_src_target[1].lower()),
                                 'related_{}'.format(lst_src_target[0].lower()))
            self.fields['relation_type'] = autocomplete.Select2ListCreateChoiceField(
                    label='Relation type',
                    widget=ListSelect2(
                            # url='/vocabularies/autocomplete/{}{}relation/normal'.format(lst_src_target[0].lower(),
                            # lst_src_target[1].lower()),
                            url=reverse('apis:apis_vocabularies:generic_vocabularies_autocomplete', args=[
                                ''.join([lst_src_target[0].lower(), lst_src_target[1].lower(), 'relation']), 'normal']),
                            attrs=attrs))
            self.fields['target'] = autocomplete.Select2ListCreateChoiceField(
                    label=lst_src_target[1],
                    widget=ListSelect2(
                            # url='/entities/autocomplete/{}'.format(lst_src_target[1].lower()),
                            url=reverse('apis:apis_entities:generic_entities_autocomplete',
                                        args=[lst_src_target[1].lower()]),
                            attrs=attrs_target),
                    validators=[validate_target_autocomplete],
                    help_text=help_text_target)
        elif entity_type.lower() == lst_src_target[1].lower():
            self.rel_accessor = (lst_src_target[0], False,
                                 'related_{}'.format(lst_src_target[0].lower()),
                                 'related_{}'.format(lst_src_target[1].lower()))
            self.fields['relation_type'] = autocomplete.Select2ListCreateChoiceField(
                    label='Relation type',
                    widget=ListSelect2(
                            url=reverse('apis:apis_vocabularies:generic_vocabularies_autocomplete', args=[
                                ''.join([lst_src_target[0].lower(), lst_src_target[1].lower(), 'relation']),
                                'reverse']),
                            attrs=attrs))
            self.fields['target'] = autocomplete.Select2ListCreateChoiceField(
                    label=lst_src_target[0],
                    widget=ListSelect2(
                            # url='/entities/autocomplete/{}'.format(lst_src_target[0].lower()),
                            url=reverse('apis:apis_entities:generic_entities_autocomplete',
                                        args=[lst_src_target[0].lower()]),
                            attrs=attrs_target),
                    validators=[validate_target_autocomplete],
                    help_text=help_text_target)
        else:
            print('no hit rel_accessor')
        if instance and instance.id:
            self.fields['target'].choices = [
                (str(Uri.objects.filter(entity=getattr(instance, self.rel_accessor[2]))[0]),
                 str(getattr(instance, self.rel_accessor[2])))]
            self.fields['target'].initial = (str(Uri.objects.filter(entity=getattr(instance, self.rel_accessor[2]))[0]),
                                             str(getattr(instance, self.rel_accessor[2])))
            if self.rel_accessor[1]:
                self.fields['relation_type'].choices = [(instance.relation_type.id,
                                                         instance.relation_type.label)]
                self.fields['relation_type'].initial = (instance.relation_type.id, instance.relation_type.label)
            else:
                self.fields['relation_type'].choices = [(instance.relation_type.id,
                                                         instance.relation_type.label_reverse)]
                self.fields['relation_type'].initial = (instance.relation_type.id, instance.relation_type.label_reverse)
        if highlighter:
            css_notes = 'HL'
            log.warning("highlighter was true")
        else:
            log.warning("highlighter was false")

        if self.relation_form.__name__.lower() == "personinstitution":
            self.fields["write citation"] = forms.BooleanField()
            self.fields["write citation"].initial = True
            self.fields["write citation"].required = False
            self.helper.include_media = False
            self.helper.layout = Layout(
            'relation_type',
            'target',
            'start_date_written',
            'end_date_written',
            'write citation',
            Accordion(
                    AccordionGroup(
                            'Notes and References',
                            'notes',
                            'references',
                            active=False,
                            css_id="{}_{}_notes_refs".format(self.relation_form.__name__, css_notes))))
        else:
            self.helper.include_media = False
            self.helper.layout = Layout(
                    'relation_type',
                    'target',
                    'start_date_written',
                    'end_date_written',
                    Accordion(
                            AccordionGroup(
                                    'Notes and References',
                                    'notes',
                                    'references',
                                    active=False,
                                    css_id="{}_{}_notes_refs".format(self.relation_form.__name__, css_notes))))

        if highlighter:
            self.fields['HL_start'] = forms.IntegerField(widget=forms.HiddenInput)
            self.fields['HL_end'] = forms.IntegerField(widget=forms.HiddenInput)
            self.fields['HL_text_id'] = forms.CharField(widget=forms.HiddenInput)
            self.helper.layout.extend([
                'HL_start',
                'HL_end',
                'HL_text_id'])

        if instance != None:

            if instance.start_date_written:
                self.fields['start_date_written'].help_text = DateParser.get_date_help_text_from_dates(
                        single_date=instance.start_date,
                        single_start_date=instance.start_start_date,
                        single_end_date=instance.start_end_date,
                        single_date_written=instance.start_date_written,
                )
            else:
                self.fields['start_date_written'].help_text = DateParser.get_date_help_text_default()

            if instance.end_date_written:
                self.fields['end_date_written'].help_text = DateParser.get_date_help_text_from_dates(
                        single_date=instance.end_date,
                        single_start_date=instance.end_start_date,
                        single_end_date=instance.end_end_date,
                        single_date_written=instance.end_date_written,
                )
            else:
                self.fields['end_date_written'].help_text = DateParser.get_date_help_text_default()

        if not instance or not instance.id or instance.id == "None":
            if prefill_inst:
                if self.fields.get("target"):
                    self.fields['target'].choices = [
                        (str(Uri.objects.filter(entity=prefill_inst)[0]),
                         str(prefill_inst))]
                    self.fields['target'].initial = (str(Uri.objects.filter(entity=prefill_inst)[0]),
                                                     str(prefill_inst))
            if prefill_func:
                if self.fields.get("relation_type"):
                    self.fields['relation_type'].choices = [(prefill_func.id,
                                                             prefill_func.label)]
                    self.fields['relation_type'].initial = (prefill_func.id, prefill_func.label)



import django_tables2 as tables
from django_tables2.utils import A
from django.conf import settings
from django.db.models import Case, When

from apis_core.apis_labels.models import Label
from apis_core.apis_metainfo.models import Uri
from apis_core.apis_metainfo.tables import (
    generic_order_start_date_written,
    generic_order_end_date_written,
    generic_render_start_date_written,
    generic_render_end_date_written
)
from apis_core.apis_relations.models import AbstractRelation



empty_text_default = 'There are currently no relations'



def get_generic_relation_listview_table(relation_name):
    """
    Creates a table class according to the relation class given by the relation_name parameter.
    Instantiates corresponding columns which also provide linking to the respectively related entities.

    :param relation_name: str : The name of the relation class to be loaded
    :return: a django-tables2 Table Class tailored for the respective relation class
    """


    # create all variables which save the foreign key fields which are different for each relation class
    relation_class = AbstractRelation.get_relation_class_of_name(relation_name)
    related_entity_class_name_a = relation_class.get_related_entity_classA().__name__.lower()
    related_entity_class_name_b = relation_class.get_related_entity_classB().__name__.lower()
    related_entity_field_name_a = relation_class.get_related_entity_field_nameA()
    related_entity_field_name_b = relation_class.get_related_entity_field_nameB()

    class GenericRelationListViewTable(tables.Table):

        # reuse the logic for ordering and rendering *_date_written
        # Important: The names of these class variables must correspond to the column field name,
        # e.g. for start_date_written, the methods must be named order_start_date_written and render_start_date_written
        order_start_date_written = generic_order_start_date_written
        order_end_date_written = generic_order_end_date_written
        render_start_date_written = generic_render_start_date_written
        render_end_date_written = generic_render_end_date_written


        class Meta:
            model = relation_class

            # the fields list also serves as the defining order of them, as to avoid duplicated definitions
            fields = [
                related_entity_field_name_a,
                related_entity_field_name_b,
                'relation_type',
                'start_date_written',
                'end_date_written',
            ]
            # reuse the list for ordering
            sequence = tuple(fields)

            # This attrs dictionary I took over from the tables implementation before. No idea if and where it would be needed.
            attrs = {"class": "table table-sm table-hover table-striped table-condensed"}

        def __init__(self, *args, **kwargs):

            # LinkColumn objects provied hyperlinking to the related entities
            self.base_columns[related_entity_field_name_a] = tables.LinkColumn(
                # which url to use:
                'apis:apis_entities:generic_entities_detail_view',
                args=[
                    # which entity sub-url to load from:
                    related_entity_class_name_a,
                    # which instance identifier to use:
                    A(related_entity_field_name_a + ".pk")
                ]
            )

            # same as above
            self.base_columns[related_entity_field_name_b] = tables.LinkColumn(
                'apis:apis_entities:generic_entities_detail_view',
                args=[
                    related_entity_class_name_b,
                    A(related_entity_field_name_b + ".pk")
                ]
            )

            super().__init__(*args, **kwargs)


    return GenericRelationListViewTable



def get_generic_relations_table(relation_class, entity_instance, detail=None):
    """
    Creates a table class according to the relation and entity class given by the parameters.

    :param relation_class: the class where the entity_instance can have instantiated relations to
    :param entity_instance: the entity instance of which related relations and entities are to be displayed
    :param detail: boolean : if this Table is to be displayed in an detail or edit UI
    :return: a django-tables2 Table Class tailored for the respective relation class and entity instance
    """
    # create all variables which save the foreign key fields which are different for each relation class
    entity_class_name = entity_instance.__class__.__name__.lower()
    related_entity_class_name_a = relation_class.get_related_entity_classA().__name__.lower()
    related_entity_class_name_b = relation_class.get_related_entity_classB().__name__.lower()
    related_entity_field_name_a = relation_class.get_related_entity_field_nameA()
    related_entity_field_name_b = relation_class.get_related_entity_field_nameB()

    # find out what other entity class the current entity instance in a given relation class is related to
    # (needed for linkg towards instances of related entities)
    if entity_class_name == related_entity_class_name_a == related_entity_class_name_b:
        other_related_entity_class_name = entity_class_name

    elif entity_class_name == related_entity_class_name_a:
        other_related_entity_class_name = related_entity_class_name_b

    elif entity_class_name == related_entity_class_name_b:
        other_related_entity_class_name = related_entity_class_name_a

    else:
        raise Exception(
            "Did not find the entity instance in the given relation class fields!" +
            "Either a wrong entity instance or wrong relation class was passed to this function."
        )


    class RelationTableBase(tables.Table):
        """
        The base table from which detail or edit tables will inherit from in order to avoid redundant definitions
        """

        # reuse the logic for ordering and rendering *_date_written
        # Important: The names of these class variables must correspond to the column field name,
        # e.g. for start_date_written, the methods must be named order_start_date_written and render_start_date_written
        order_start_date_written = generic_order_start_date_written
        order_end_date_written = generic_order_end_date_written
        render_start_date_written = generic_render_start_date_written
        render_end_date_written = generic_render_end_date_written

        class Meta:
            """
            Meta class needed for django-tables2 plugin.
            """

            empty_text = empty_text_default
            model = relation_class

            # the fields list also serves as the defining order of them, as to avoid duplicated definitions
            fields = [
                'start_date_written',
                'end_date_written',
                'other_relation_type',
                "other_related_entity"
            ]
            # reuse the list for ordering
            sequence = tuple(fields)

            # This attrs dictionary I took over from the tables implementation before. No idea if and where it would be needed.
            attrs = {
                "class": "table table-sm table-hover table-striped table-condensed",
                "id": related_entity_class_name_a.title()[:2] + related_entity_class_name_b.title()[:2] + "_conn"
            }

        def render_other_related_entity(self, record, value):
            """
            Custom render_FOO method for related entity linking. Since the 'other_related_entity' is a generated annotation
            on the queryset, it does not return the related instance but only the foreign key as the integer it is.
            Thus fetching the related instance is necessary.

            :param record: The 'row' of a queryset, i.e. an entity instance
            :param value: The current column of the row, i.e. the 'other_related_entity' annotation
            :return: related instance
            """

            if value == record.get_related_entity_instanceA().pk :
                return record.get_related_entity_instanceA()

            elif value == record.get_related_entity_instanceB().pk :
                return record.get_related_entity_instanceB()

            else:
                raise Exception(
                    "Did not find the entity this relation is supposed to come from!" +
                    "Something must have went wrong when annotating for the related instance."
                )


        def __init__(self, data, *args, **kwargs):

            # annotations for displaying data about the 'other side' of the relation.
            # Both of them ('other_related_entity' and 'other_relation_type') are necessary for displaying relations
            # in context to what entity we are calling this from.
            data = data.annotate(
                # In order to provide the 'other instance' for each instance of a table where this whole logic is called from,
                # the queryset must be annotated accordingly. The following Case searches which of the two related instances
                # of a relation queryset entry is the one corresponding to the current entity instance. When found, take the
                # other related entity (since this is the one we are interested in displaying).
                #
                # The approach of using queryset's annotate method allows for per-instance case decision and thus
                # guarantees that the other related entity is always correctly picked,
                # even in case two entities are of the same class.
                other_related_entity=Case(
                    # **kwargs pattern is needed here as the key-value pairs change with each relation class and entity instance.
                    When(**{
                        related_entity_field_name_a + "__pk": entity_instance.pk,
                        "then": related_entity_field_name_b
                    }),
                    When(**{
                        related_entity_field_name_b + "__pk": entity_instance.pk,
                        "then": related_entity_field_name_a
                    }),
                )
            ).annotate(
                # Get the correct side of the relation type given the current entity instance.
                #
                # The approach of using queryset's annotate method allows for per-instance case decision and thus
                # guarantees that the other related entity is always correctly picked,
                # even in case two entities are of the same class.
                other_relation_type=Case(
                When(**{
                    # A->B relation and current entity instance is A, hence take forward name
                    related_entity_field_name_a + "__pk": entity_instance.pk,
                    "then": "relation_type__name"
                }),
                When(**{
                    # A->B relation and current entity instance is B, hence take reverse name.
                    related_entity_field_name_b + "__pk": entity_instance.pk,
                    "then": "relation_type__name_reverse"
                }),
            )
            )
            for an in data:
                if getattr(an, f"{related_entity_field_name_a}_id") == entity_instance.pk:
                    an.other_relation_type = getattr(an.relation_type, "label")
                else:
                    an.other_relation_type = getattr(an.relation_type, "label_reverse")


            super().__init__(data, *args, **kwargs)


    if detail:

        class RelationTableDetail(RelationTableBase):
            """
            Sublcass inheriting the bulk of logic from parent. This table is used for the 'detail' views.
            """

            def __init__(self, data, *args, **kwargs):

                # Only addition with respect to parent class is which main url is to be used when clicking on a
                # related entity column.
                self.base_columns["other_related_entity"] = tables.LinkColumn(
                    'apis:apis_entities:generic_entities_detail_view',
                    args=[
                        other_related_entity_class_name,
                        A("other_related_entity")
                    ],
                    verbose_name="Related " + other_related_entity_class_name.title()
                )

                super().__init__(data=data, *args, **kwargs)


        return RelationTableDetail


    else:

        class RelationTableEdit(RelationTableBase):
            """
            Sublcass inheriting the bulk of logic from parent. This table is used for the 'edit' view.
            """

            class Meta(RelationTableBase.Meta):
                """
                Additional Meta fields are necessary for editing functionalities
                """

                # This fields list also defines the order of the elements.

                if relation_class.__name__ == "PersonInstitution" and entity_class_name == "person":
                    fields = ["delete"] + RelationTableBase.Meta.fields + ["edit"] + ["autoDate"]
                else:
                    fields = ["delete"] + RelationTableBase.Meta.fields + ["edit"]

                if 'apis_bibsonomy' in settings.INSTALLED_APPS:
                    fields = ["ref"] + fields

                # again reuse the fields list for ordering
                sequence = tuple(fields)


            def __init__(self, *args, **kwargs):

                # Clicking on a related entity will lead also the edit view of the related entity instance
                self.base_columns["other_related_entity"] = tables.LinkColumn(
                    'apis:apis_entities:generic_entities_edit_view',
                    args=[
                        other_related_entity_class_name, A("other_related_entity")
                    ],
                    verbose_name="Related " + other_related_entity_class_name.title()
                )

                # delete button
                self.base_columns['delete'] = tables.TemplateColumn(
                    template_name='apis_relations/delete_button_generic_ajax_form.html'
                )

                # edit button
                self.base_columns['edit'] = tables.TemplateColumn(
                    template_name='apis_relations/edit_button_generic_ajax_form.html'
                )

                # autoDate button
                if relation_class.__name__ in ["PersonInstitution", "PersonPlace"]:# and entity_class_name == "person":
                    self.base_columns['autoDate'] = tables.TemplateColumn(
                        template_name = "element_templates/autoDate_button_generic_ajax_form.html"
                    )

                

                # bibsonomy button
                if 'apis_bibsonomy' in settings.INSTALLED_APPS:
                    self.base_columns['ref'] = tables.TemplateColumn(
                        template_name='apis_relations/references_button_generic_ajax_form.html'
                    )

                super().__init__(*args, **kwargs)


        return RelationTableEdit



class EntityUriTable(tables.Table):

    delete = tables.TemplateColumn(template_name='apis_relations/delete_button_Uri_ajax_form.html')

    class Meta:
        empty_text = empty_text_default
        model = Uri
        fields = ['uri']
        sequence = ('delete', 'uri')
        # add class="paleblue" to <table> tag
        attrs = {"class": "table table-hover table-striped table-condensed",
                 "id": "PURI_conn"}



class LabelTableBase(tables.Table):

    label2 = tables.TemplateColumn(template_name="apis_relations/labels_label.html")

    # reuse the logic for ordering and rendering *_date_written
    # Important: The names of these class variables must correspond to the column field name,
    # e.g. for start_date_written, the methods must be named order_start_date_written and render_start_date_written
    order_start_date_written = generic_order_start_date_written
    order_end_date_written = generic_order_end_date_written
    render_start_date_written = generic_render_start_date_written
    render_end_date_written = generic_render_end_date_written

    class Meta:

        empty_text = empty_text_default
        model = Label

        # Note that as the next attribute 'sequence' builds on this list 'fields', the order defined within this list
        # will be reused for the tuple 'sequence'. So if the order needs to be changed, better do it here in the list 'fields'.
        fields = ['start_date_written', 'end_date_written', 'label_type', 'isoCode_639_3']
        sequence = ('label2',) + tuple(fields)

        # add class="paleblue" to <table> tag
        attrs = {
            "class": "table table-sm table-hover table-striped table-condensed",
            "id": "PL_conn"
        }



class LabelTableEdit(LabelTableBase):
    """
    Reuse most of the base table class for labels. Only addition is editing functionality.
    """

    edit = tables.TemplateColumn(template_name='apis_relations/edit_button_persLabel_ajax_form.html')

    class Meta(LabelTableBase.Meta):
        sequence = LabelTableBase.Meta.sequence + ("edit",)



def get_generic_entity_progress_table(data, version=True):

    class GenericEntityProgressTable(tables.Table):
        if version:
            type = tables.columns.TemplateColumn(template_code=u"""{{record.version}}""", default="")
        name = tables.columns.TemplateColumn(template_code=u"""<a href='#' onclick='updateSessionEntity("{{record.entity_type}}","{{record.id}}")'> {{record.name}}</a>""", default="")
        start = tables.columns.TemplateColumn(template_code=u"""{{record.start_date_written}}""", default="")
        end = tables.columns.TemplateColumn(template_code=u"""{{record.end_date_written}}""", default="")
        labels = tables.columns.TemplateColumn(template_code=u"""{{record.labels}}""", default="")
        if "titles" in data[0].keys():
            titles = tables.columns.TemplateColumn(template_code=u"""{{record.titles}}""", default="")
        if "gender" in data[0].keys():
            gender = tables.columns.TemplateColumn(template_code=u"""{{record.gender}}""", default="")
        if "first_name" in data[0].keys():
            first_name = tables.columns.TemplateColumn(template_code=u"""{{record.first_name}}""", default="")
        if "kind" in data[0].keys():
            kind = tables.columns.TemplateColumn(template_code=u"""{{record.kind}}""", default="")


        class Meta:
            attrs = {"class": "table table-sm table-hover table-striped table-condensed pb-5"}
            fields = ["name", "start", "end", "labels"]
            if "first_name" in data[0].keys():
                fields.insert(1, "first_name")
            elif "kind" in data[0].keys():
                fields.insert(1, "kind")
            if "titles" in data[0].keys():
                fields += ["titles"]
            if "gender" in data[0].keys():
                fields.insert(4, "gender")
            if version:
                fields.insert(0, "type")
            sequence = fields

    return GenericEntityProgressTable(data)

def get_generic_relation_progress_table(data, version=True):

    A_class = data[0]["relation_source"]["class"]
    B_class = data[0]["relation_target"]["class"]

    if A_class == B_class:
        A_class += "A"
        B_class += "B"



    class GenericRelationProgressTable(tables.Table):
        if version:
            type = tables.columns.TemplateColumn(template_code=u"""{{record.version}}""", default="")
        start = tables.columns.TemplateColumn(template_code=u"""{{record.start_date_written}}""", default="")
        end = tables.columns.TemplateColumn(template_code=u"""{{record.end_date_written}}""", default="")
        objectA = tables.columns.TemplateColumn(template_code=u"""<a href='#' onclick='updateSessionEntity("{{record.relation_source.class}}","{{record.relation_source.id}}")'> {{record.relation_source.name}}</a>""", verbose_name=A_class, default="")
        relation = tables.columns.TemplateColumn(template_code=u"""{{record.relation_type.name}}""", default="")
        objectB = tables.columns.TemplateColumn(template_code=u"""<a href='#' onclick='updateSessionEntity("{{record.relation_target.class}}","{{record.relation_target.id}}")'> {{record.relation_target.name}}</a>""", verbose_name=B_class, default="")

        class Meta:
            attrs = {
                "class": "table table-sm table-hover table-striped table-condensed pb-5"
            }
            fields = ["start", "end", "objectA", "relation", "objectB"]
            if version:
                fields.insert(0, "type")
            sequence = fields

    return GenericRelationProgressTable(data)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apis_import_project', 'apis_import_project.migrations']

package_data = \
{'': ['*'],
 'apis_import_project': ['static/apis_import_project/css/*',
                         'static/apis_import_project/js/*',
                         'templates/element_templates/*',
                         'templates/pages/*',
                         'templates/section_templates/*']}

setup_kwargs = {
    'name': 'apis-import-project',
    'version': '0.3.4',
    'description': 'Generic Django-App for APIS-instances to support importing new data manually with a streamlined workflow and some automation.',
    'long_description': None,
    'author': 'Gregor Pirgie',
    'author_email': 'gregor.pirgie@oeaw.ac.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
